from datetime import datetime

from typing import Dict

from pydantic import BaseModel, Field, Extra, root_validator

from chaiverse.config import get_elo_base_rating, get_elo_base_submission_id
from chaiverse.lib.elo import get_elo_ratings, get_mle_elo_scores
from chaiverse.lib.now import utcnow


class PreferenceSummary(BaseModel):
    '''
    wins_dict[player_a,player_b] is a sparse matrix of win count for player a vs player b
    elo_scores[player] must include all players, and the value is last known best elo score
    '''
    wins_dict: Dict[str, Dict[str, int]] = Field(default={})
    elo_scores: Dict[str, float] = Field(default={})

    @root_validator
    def validate_elo_scores(cls, value):
        wins_dict = value.get('wins_dict')
        initial_elo_scores = value.get('elo_scores')
        value['elo_scores'] = get_mle_elo_scores(initial_elo_scores, wins_dict)
        return value

    def wins(self, a_id: str, b_id: str):
        wins = self.wins_dict.get(a_id, {}).get(b_id, 0)
        return wins

    def rounds(self, a_id: str, b_id: str):
        return self.wins(a_id, b_id) + self.wins(b_id, a_id)

    def win_rate(self, a_id: str, b_id: str):
        return self.wins(a_id, b_id) / self.rounds(a_id, b_id)

    def calculate_elo_ratings(self, elo_base_id: str = None, elo_base_rating: float = None):
        elo_ratings = {}
        if self.elo_scores:
            elo_base_id = elo_base_id or get_elo_base_submission_id()
            elo_base_rating = elo_base_rating or get_elo_base_rating()
            elo_base_score = self.elo_scores[elo_base_id]
            elo_ratings = get_elo_ratings(self.elo_scores, elo_base_score, elo_base_rating)
        return elo_ratings


class PreferenceSummaries(BaseModel, extra=Extra.allow):
    summaries: Dict[str, PreferenceSummary] = Field(default={})
    created_at: datetime = Field(default_factory=utcnow)



from base64 import b64encode, b64decode
import dill

def serialize_function(function):
    # Pickle the function (use dill as usual
    # pickle doesn't work on locals/lambdas
    function = dill.dumps(function)
    # Convert to JSON-serialisable byte string
    function = b64encode(function)
    # Convert to plain string
    function = function.decode("utf-8")
    return function

def deserialize_function(function):
    function = b64decode(function)
    function = dill.loads(function)
    return function



from datetime import datetime
from typing import Optional

from pydantic import BaseModel, root_validator, validator


class DateRange(BaseModel):
    start_date: Optional[datetime]
    end_date: Optional[datetime]

    @validator('start_date', 'end_date')
    def validate_date(cls, value):
        assert value is None or value.tzinfo, 'date is missing time zone'
        return value

    @root_validator
    def validate_start_date_and_end_date(cls, values):
        start_date = values.get('start_date')
        end_date = values.get('end_date')
        if start_date and end_date:
            assert start_date <= end_date, 'date_range start date is after end_date'
        return values

    @property
    def start_epoch_time(self):
        epoch_time = self.start_date.timestamp() if self.start_date else 0
        return epoch_time

    @property
    def end_epoch_time(self):
        epoch_time = self.end_date.timestamp() if self.end_date else float('inf')
        return epoch_time


class BoundedDateRange(DateRange):
    @root_validator
    def validate_start_date_and_end_date(cls, values):
        start_date = values.get('start_date')
        end_date = values.get('end_date')
        assert start_date, 'bounded date_range missing start_date'
        assert end_date, 'bounded date_range missing end_date'
        # https://github.com/pydantic/pydantic/pull/2670 pydantic bug fixed in 2.6 but we are not in that version
        # since pydantic 2.6, parent root validator will not be called if overriden
        assert start_date <= end_date, 'date_range start_date is after end_date'
        return values



from datetime import datetime
from typing import List

import numpy as np
import pandas as pd
from pydantic import BaseModel, Extra, Field

from chaiverse import config
from chaiverse.lib.now import utcnow
from chaiverse.schemas.leaderboard_row_schema import LeaderboardRow


DEFAULT_LEADERBOARD_HEADER_MAPPING = {
    'double_thumbs_up_ratio': '👍👍_ratio',
    'thumbs_up_ratio': '👍👍+👍_ratio',
    'single_thumbs_up_ratio': '👍_ratio',
    'thumbs_down_ratio': '👎_ratio',
}

DEFAULT_LEADERBOARD_INCLUDES = [
    'developer_uid',
    'submission_id',
    'elo_rating',
    'win_ratio',
    'num_battles',
    'double_thumbs_up_ratio',
    'thumbs_down_ratio',
    'feedback_count',
    'model_score',
    'safety_score',
    'best_of',
    'max_input_tokens',
    'model_repo',
    'reward_repo',
    'status',
    'model_name',
    'us_pacific_date',
]

DEFAULT_LEADERBOARD_EXCLUDES = []

DEFAULT_LEADERBOARD_SORT_PARAMS = {
    'by': 'elo_rating',
    'ascending': False
}

DEFAULT_TABULATE_OPTIIONS = {
    'numalign': 'decimal',
}


class Leaderboard(BaseModel, extra=Extra.allow):
    timestamp: datetime = Field(default_factory=lambda: utcnow())
    leaderboard_rows: List[LeaderboardRow]

    @property
    def df(self) -> pd.DataFrame:
        leaderboard_rows = [row.all_fields_dict() for row in self.leaderboard_rows]
        df = pd.DataFrame.from_records(leaderboard_rows)
        df = _sort_by_values(df, **DEFAULT_LEADERBOARD_SORT_PARAMS)
        return df

    def to_display_df(self, includes=None, excludes=None, sort_params=None, header_mapping=None) -> pd.DataFrame:
        includes = includes or DEFAULT_LEADERBOARD_INCLUDES
        excludes = excludes or DEFAULT_LEADERBOARD_EXCLUDES
        sort_params = sort_params or DEFAULT_LEADERBOARD_SORT_PARAMS
        header_mapping = header_mapping or DEFAULT_LEADERBOARD_HEADER_MAPPING
        df = self.df
        if len(df) > 0:
            df = _add_repo_aggregate_columns(df)
            df = _include_listed_columns(df, includes)
            df = _exclude_listed_columns(df, excludes)
            df = _sort_by_values(df, **sort_params)
            df['elo_rating'] = df.elo_rating.astype(int)
            df = df.round(2)
            df = df.rename(columns=header_mapping)
        return df

    def to_html(self, includes=None, header_mapping=None):
        includes = includes or DEFAULT_LEADERBOARD_INCLUDES
        df = self.to_display_df(includes=includes, header_mapping=header_mapping)
        df.fillna("", inplace=True)
        html = df.to_html(classes="leaderboard-table table display nowrap", index=False, justify="center", escape=False)
        return html

    @property
    def auto_deactivation_candidates(self):
        rows = _get_can_auto_deactivate_rows(self.leaderboard_rows)
        rows = _sort_auto_deactivate_rows(rows)
        rows = _remove_top_models(rows)
        submission_ids = [row.submission_id for row in rows]
        return submission_ids


def _add_repo_aggregate_columns(df):
    for column in ["celo_rating", "elo_rating", "double_thumbs_up_ratio", "thumbs_down_ratio", "win_ratio", "safety_score", "model_score"]:
        df[f"{column}_by_repo"] = df.groupby("model_repo")[column].transform("max")
    for column in ["feedback_count", "num_battles"]:
        df[f"{column}_by_repo"] = df.groupby("model_repo")[column].transform("sum")
    return df


def _include_listed_columns(df, includes):
    df = df[[column for column in includes if column in df.columns]]
    return df


def _exclude_listed_columns(df, excludes):
    df = df[[column for column in df.columns if column not in excludes]]
    return df


def _sort_by_values(df, by: List[str], ascending: bool):
    if len(df) > 0:
        df = df.sort_values(by=by, ascending=ascending, na_position='last', ignore_index=True)
        df.index = np.arange(1, len(df) + 1)
    return df


def _get_can_auto_deactivate_rows(rows: List[LeaderboardRow]):
    rows = [row for row in rows if row.can_auto_deactivate()]
    return rows


def _sort_auto_deactivate_rows(rows: List[LeaderboardRow]):
    rows = sorted(rows, key=lambda row: row.elo_rating, reverse=True)
    return rows


def _remove_top_models(rows: List[LeaderboardRow]):
    rows = rows[config.AUTO_DEACTIVATION_MIN_RANK - 1:]
    return rows
